import subprocess
import os
import logging

def extract_rar_file(rar_filepath, tmp_dir):
    """ Extracts files from .rar file located in rar_filepath to tmp_dir.
    """
    logging.info('Extracting {}...'.format(rar_filepath))
    return subprocess.run(["rar", "x", rar_filepath, tmp_dir])

def check_rar_exists():
    try:
        p = subprocess.run(["rar"], stdout=subprocess.PIPE)
        return True
    except FileNotFoundError:
        return False



import os
#field which is used to store meta information about the file in ElasticSearch, e.g. path, extension, type
META_FIELD = "properties"
ES_URL = os.getenv("DOCPARSER_ES_URL", "http://localhost:9200")
BROKER_URL = "redis://localhost:6379/0"
RESULT_BACKEND_URL = "redis://localhost:6379/0"

FACE_VECTOR_FIELD_NAME = "texta_face_vectors"



import logging
import mailbox
import multiprocessing
import os
import shutil
from random import getrandbits
from typing import Optional

import magic

from texta_parsers.tools.message import AttachmentParser, BodyParser, HeaderParser
from texta_parsers.tools.pst_extractor import check_readpst_exists, extract_pst_file


SUPPORTED_MAIL_FORMATS = [".mbox", ".pst", ".eml"]
logging.basicConfig(format='%(levelname)s %(asctime)s: %(message)s',
                    datefmt='%d.%m.%Y %H:%M:%S',
                    level=logging.INFO)


class EmailParser():
    """ 
    # Parameters 
    max_attachment_size (int): in megabytes. Content from larger attachments won't be parsed (optional, default: 10)
    num_threads (int): number of threads in case of multiprocessing, (optional, default: 1)
    tmp_folder (string): folder for temporal files, (optional, default: "tmp")
    save_path (string): folder for non-temporal files, (optional, default: "parsed_files")
    """

    def __init__(self, num_threads=1, tmp_folder="tmp", save_path="parsed_files"):
        self.num_threads = num_threads
        self.tmp_folder = tmp_folder
        self.save_path = save_path
        if not check_readpst_exists():
            raise FileNotFoundError("Readpst not found. Can not initialize parser.")


    def parse(self, input_path: str, save_mails: bool = False):
        """Parses emails from given folder or file.

        # Parameters
        input_path (str): Path to folder or file that contains emails
        save_mails (bool): whether to save emails to files (optional, default: False)

        # Returns
        Generator which yields emails in format (msg_json, [attachment_json])
        """
        email_files = self._get_files(input_path)
        self.save_mails = save_mails
        self._make_dirs()

        for email_file, ext in email_files:
            logging.info("Parsing {}".format(email_file))

            if ext == ".mbox":
                folder_name = email_file.split("/")[-1].split(".")[0]

                with multiprocessing.Pool(processes=self.num_threads) as pool:
                    for msg_json, attachment_jsons in pool.imap(self._parse_msg, mailbox.mbox(email_file)):
                        if (msg_json != {}):
                            msg_json["original_folder"] = folder_name
                        yield msg_json, attachment_jsons

            elif ext == ".pst":
                extracted_mboxes = extract_pst_file(email_file, self.tmp_folder)
                for mbox_file in extracted_mboxes:
                    logging.info("Parsing from .pst {}".format(mbox_file))
                    folder_name = mbox_file.split("/")[-1].split(".")[0]

                    with multiprocessing.Pool(processes=self.num_threads) as pool:
                        for msg_json, attachment_jsons in pool.imap(self._parse_msg, mailbox.mbox(mbox_file)):
                            if (msg_json != {}):
                                msg_json["original_folder"] = folder_name
                            yield msg_json, attachment_jsons

                self._delete_extracted_pst(self.tmp_folder)

            elif ext == ".eml":
                raw_msg = self._load_raw_mail(email_file)
                msg = mailbox.Message(raw_msg)
                msg_json, attachment_jsons = self._parse_msg(msg)
                yield msg_json, attachment_jsons


    def _parse_msg_rec(self, msg, msg_id):
        """Parses given message recursively using parsers in message module for different parts
        """
        is_root_msg = msg_id == ""

        if is_root_msg:  # is main email, parse it accordingly
            msg_data, attachments = self._get_all_msg_data(msg)
            msg_id = msg_data["id"]

        else:  # is message/rfc822 attachment, parse it as an email with respective parsers
            try:
                filename = AttachmentParser.parse_filename(msg.get_filename())
            except:
                filename = ""

            msg = msg.get_payload()[0]
            msg_data, attachments = self._get_all_msg_data(msg)
            msg_data["parent_id"] = msg_id
            msg_data["filename"] = filename

            for attachment in attachments:
                #replace parent_id the id of correct parent
                attachment["parent_id"] = msg_data["id"]
            attachments.append(msg_data)

        # parse other message/rfc822 attachments if there is any
        if msg.is_multipart():
            for payload in msg.get_payload():
                if payload.get_content_type() == "message/rfc822":
                    _, nested_attachments = self._parse_msg_rec(payload, msg_id)

                    if (self.save_mails):
                        #save sub-mails if required as we won't see it's payload later
                        print("saving mail with id", nested_attachments[-1]["id"])
                        filepath = self._save_mail_as_eml(payload, msg_id + "/" + nested_attachments[-1]["id"])
                        if (filepath):
                            nested_attachments[-1]["location"] = filepath

                    attachments += nested_attachments

        return msg_data, attachments


    def _parse_msg(self, msg):
        try:
            msg_json, attachment_jsons = self._parse_msg_rec(msg, "")
            if (self.save_mails):
                filepath = self._save_mail_as_eml(msg, msg_json["id"])
                if (filepath):
                    msg_json["location"] = filepath
            return msg_json, attachment_jsons
        except Exception as e:
            # in case all other more specific exceptions handlers have failed
            logging.exception(e)
            return {}, [{}]


    def _get_all_msg_data(self, msg):
        h_parser = HeaderParser(msg)
        b_parser = BodyParser(msg)

        msg_data = h_parser.parse()
        body = b_parser.parse_body()
        msg_data["body"] = body

        msg_id = str(getrandbits(64))
        msg_data["id"] = msg_id

        attachments = b_parser.parse_attachments(msg_id)

        # concatenate errors
        header_fails = h_parser.fails
        header_fails.update(b_parser.fails)

        if header_fails != {}:
            msg_data["errors"] = header_fails

        return msg_data, attachments


    @staticmethod
    def _try_get_file_beginning(file):
        if isinstance(file, bytes):
            return file[:2048]
        else:
            try:
                return open(file, "rb").read(2048)
            except:
                return None


    def _get_files(self, path):
        """Returns a list of email files from given folder (recursive) or a list with single file if input is file.
        """
        results = []
        if os.path.isfile(path):
            file_beginning = EmailParser._try_get_file_beginning(path)
            res = self._file_is_mail(path, file_beginning)
            if (res):
                results.append((path, res))
        else:
            for dirpath, _, filenames in os.walk(path):
                for filename in filenames:
                    full_file = os.path.join(dirpath, filename)

                    file_beginning = EmailParser._try_get_file_beginning(full_file)
                    res = self._file_is_mail(filename, file_beginning)
                    if (res):
                        results.append((full_file, res))

        return results


    @staticmethod
    def _file_is_mail(filename, file_beginning):
        ext = os.path.splitext(filename)[1]
        if ext in SUPPORTED_MAIL_FORMATS:
            return ext

        if (file_beginning == None):
            return
        elif (EmailParser._has_msg_mimetype(file_beginning)):
            return ".eml"
        elif (EmailParser._begins_like_mbox(file_beginning)):
            return ".mbox"


    @staticmethod
    def _has_msg_mimetype(input):
        mimetype = magic.from_buffer(input, mime=True)
        return mimetype == "message/rfc822"


    @staticmethod
    def _begins_like_mbox(input):
        try:
            return input.decode().startswith("From ")
        except:
            return False


    def _load_raw_mail(self, path):
        with open(path, "rb") as f:
            raw_mail = f.read()
        return raw_mail


    def _make_dirs(self):
        for path in [self.tmp_folder, self.save_path]:
            if (path != None):
                if not os.path.exists(path):
                    os.makedirs(path)


    def _save_mail_as_eml(self, message, msg_id):
        save_path = os.path.join(self.save_path, "mails")
        msg_path = os.path.join(save_path,  msg_id + ".eml")

        if not os.path.exists(os.path.dirname(msg_path)):
            os.makedirs(os.path.dirname(msg_path))

        try:
            with open(msg_path, "wb") as f:
                f.write(bytes(message))
            return msg_path
        except KeyError:
            logging.error("Failed flattening file with id {}".format(msg_id))
            return None


    def _delete_extracted_pst(self, pst_dir):
        """Deletes the content of the tmp folder which is used for temporarily extracting pst files
        """
        for root, dirs, files in os.walk(pst_dir):
            for f in files:
                os.unlink(os.path.join(root, f))
            for d in dirs:
                shutil.rmtree(os.path.join(root, d))



import subprocess
import os
import logging

def extract_tar_file(tar_filepath, tmp_dir):
    """ Extracts files from .tar file located in tar_filepath to tmp_dir.
    """
    logging.info('Extracting {}...'.format(tar_filepath))
    return subprocess.run(["tar", "-xf", tar_filepath, "-C", tmp_dir])

def check_tar_exists():
    try:
        p = subprocess.run(["tar"], stdout=subprocess.PIPE)
        return True
    except FileNotFoundError:
        return False



import logging

logging.basicConfig(
    format='%(levelname)s %(asctime)s: %(message)s',
    datefmt='%d.%m.%Y %H:%M:%S',
    level=logging.INFO
)


class MLPProcessor:
    """
    A wrapper class to apply MLP to DocParser documents.
    """
    def __init__(self, mlp):
        self.mlp = mlp
        self.IGNORE_FIELDS = ("texta_facts", "properties", "texta_face_vectors")


    def apply_mlp(self, generator, analyzers=["all"]):
        """ Applies MLP to objects in a given generator.
        """
        for item in generator:
            # check if email (it returns a tuple because of attachments)
            if isinstance(item, tuple):
                email, attachments = item
                self._apply_mlp_to_mails(email, attachments, analyzers)
            else:
                self._apply_mlp_to_item(item, analyzers)
            yield item


    def _apply_mlp(self, document: dict, field: str, analyzers: list):
        if (field not in document):
            return
        content = document.get(field, "")

        if content:
            mlp_res = self.mlp.process(content, analyzers=analyzers)
            mlp_res_path = field + "_mlp"
            # Add the MLP output dictionary.
            document[mlp_res_path] = mlp_res["text"] 

            facts = []
            for f in mlp_res["texta_facts"]:
                f["doc_path"] = f"{mlp_res_path}.text"
                facts.append(f)

            if facts:
                document["texta_facts"] = facts


    def _apply_mlp_to_mails(self, email: dict, attachments: list, analyzers: list):
        self._apply_mlp(email, "body", analyzers)
        for attachment in attachments:
            self._apply_mlp(attachment, "content", analyzers)


    def _apply_mlp_to_item(self, item: dict, analyzers: list):
        # apply it to all fields as we don't know anything about the item or it's fields
        item_copy = item.copy()
        for key in item_copy.keys():
            if key not in self.IGNORE_FIELDS:
                self._apply_mlp(item, key, analyzers)

